import asyncio
import logging

from threading import Lock
from traceback import format_exc

from paramiko import RSAKey
from sshtunnel import SSHTunnelForwarder

from ._shared import (
    MessageType,
    SpawnTunnelRequest,
    SpawnTunnelResponse,
    asdict, load_frame, dump_frame, start_server
)


logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%d.%m.%Y %H:%M:%S'
)
logger = logging.getLogger()


spawn_lock = Lock()
spawned: 'dict[tuple, SSHTunnelForwarder]' = {}


def _parse_addr_string(addr: str) -> 'tuple[str] | tuple[str, int]':
    host, _, port = addr.strip().partition(':')
    if not port:
        return (host,)
    if not port.isdigit():
        raise ValueError('Incorrect address: port is not number')
    return (host, int(port))


def blocking_spawn(data: SpawnTunnelRequest):
    with spawn_lock:
        forwarder_id = (data.ssh_addr, data.ssh_user)
        if spawned_forwarder := spawned.get(forwarder_id):
            return spawned_forwarder

        key = RSAKey.from_private_key_file(data.ssh_identity_file)
        forwarder = SSHTunnelForwarder(
            ssh_address_or_host=_parse_addr_string(data.ssh_addr),
            ssh_username=data.ssh_user,
            ssh_pkey=key,
            remote_bind_address=_parse_addr_string(data.remote_addr),
            local_bind_address=('localhost',)
        )
        forwarder.start()

        logger.info('New forwarder spawned (ssh_addr=%s, remote_addr=%s)',
                    data.ssh_addr, data.remote_addr)

        spawned[forwarder_id] = forwarder
        return forwarder


async def spawn_tunnel(data: SpawnTunnelRequest):
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, blocking_spawn, data)


async def handle_request(req_type: MessageType, data):
    if req_type == MessageType.SPAWN_TUNNEL:
        data = SpawnTunnelRequest(**data)
        forwarder = await spawn_tunnel(data)
        host, port = forwarder.local_bind_address
        response = SpawnTunnelResponse(host, port)
        return (MessageType.OK, asdict(response))

    else:
        raise ValueError(f'Unknown request {req_type!r}')


async def on_connect(r: asyncio.StreamReader, w: asyncio.StreamWriter):
    req_type, req_data = load_frame(await r.readline())
    try:
        resp_type, resp_data = await handle_request(req_type, req_data)
    except Exception as e:
        logger.error(format_exc())
        resp_type = MessageType.ERROR
        resp_data = {'error': repr(e)}

    w.write(dump_frame(resp_type, resp_data))
    await w.drain()
    w.close()
    await w.wait_closed()


# TODO: reopen tunnels after service restart
async def main():
    server = await start_server(on_connect)
    logger.info('Server started!')
    try:
        await server.serve_forever()
    except KeyboardInterrupt:
        server.close()
        await server.wait_closed()
        logger.info('Server closed.')



import argparse

parser = argparse.ArgumentParser()

parser.add_argument(
    '--systemd-install',
    help='Install as systemd service',
    action='store_true',
    dest='systemd_install'
)

args = parser.parse_args()

if args.systemd_install:
    import os
    import sys

    if not os.path.exists('/run/systemd/system'):
        print('Systemd not found. '
              'This type of installation is not supported on your system.')
        sys.exit()

    template = [
        '[Unit]',
        'Description="SSH port forward daemon"',
        'After=network.target',
        '',
        '[Service]',
        'Type=simple',
       f'ExecStart="{sys.executable}" -m forwarderd',
        'Restart=always',
    ]

    UNIT_PATH = '/etc/systemd/system/forwarderd.service'

    if os.path.exists(UNIT_PATH):
        print('Unit file already exists. '
              'Use \'systemctl status forwarderd\' to check service status')
        sys.exit()

    try:
        with open(UNIT_PATH, 'w') as file:
            file.writelines(template)
    except FileNotFoundError:
        print('Systemd service directory not found. '
              'Here is text of .service file, put it in systemd '
              'service directory, then start service with systemctl:')
        print()
        print('\n'.join(template))
    else:
        print(f'Service file created at {UNIT_PATH!r}')
        print('Use \'systemctl start forwarderd\' to start daemon')

else:
    import asyncio

    from . import _daemon

    asyncio.run(_daemon.main())



from ._shared import (
    MessageType,
    SpawnTunnelRequest,
    SpawnTunnelResponse,
    asdict, load_frame, dump_frame, connect
)


async def _make_request(req_type, req_data):
    r, w = await connect()

    w.write(dump_frame(req_type, req_data))
    resp_type, resp_data = load_frame(await r.readline())

    if resp_type == MessageType.ERROR:
        raise Exception(resp_data['error'])
    elif resp_type == MessageType.OK:
        return resp_data
    else:
        raise ValueError(f'Unknown response type {resp_type!r}')


async def spawn_tunnel(data: SpawnTunnelRequest) -> SpawnTunnelResponse:
    resp = await _make_request(MessageType.SPAWN_TUNNEL, asdict(data))
    return SpawnTunnelResponse(**resp)



# pyright: reportGeneralTypeIssues=false

import os
import json
import asyncio

from enum import Enum
from dataclasses import dataclass, asdict


IP_ADDR = ('localhost', 19567)
SOCK_PATH = '/var/run/forwarderd.sock'


if hasattr(asyncio, 'open_unix_connection'):
    async def connect():
        return await asyncio.open_unix_connection(SOCK_PATH)

    async def start_server(cb):
        os.makedirs(os.path.dirname(SOCK_PATH), exist_ok=True)
        return await asyncio.start_unix_server(cb, SOCK_PATH)

else:
    async def connect():
        return await asyncio.open_connection(*IP_ADDR)

    async def start_server(cb):
        return await asyncio.start_server(cb, *IP_ADDR)


def dump_frame(type: 'MessageType', data: dict) -> bytes:
    str_dump = json.dumps([type, data], ensure_ascii=False)
    return str_dump.encode('utf-8') + b'\n'


def load_frame(data: bytes) -> 'tuple[MessageType, dict]':
    return tuple(json.loads(data[:-1]))


class MessageType(int, Enum):
    SPAWN_TUNNEL = 1

    OK = 15
    ERROR = 16


@dataclass
class SpawnTunnelRequest:
    ssh_addr: str
    ssh_user: str
    ssh_identity_file: str
    remote_addr: str


@dataclass
class SpawnTunnelResponse:
    host: str
    port: int
